local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")

function Capitalist:LinkDissect(link)
	-- Example link:
	-- "|cff9d9d9d|Hitem:7073:0:0:0:0:0:0:0:80:0|h[Broken Fang]|h|r"

	local _, _, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging, Name = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

	print(">LinkDissect")
	print(Color)
	print(Ltype)
	print(Id)
	print(Enchant)
	print(Gem1)
	print(Gem2)
	print(Gem3)
	print(Gem4)
	print(Suffix)
	print(Unique)
	print(LinkLvl)
	print(reforging)
	print(Name)
	print("<")
	print(link)
end

function Capitalist:LinkExtractId(link)
	local _, _, _, _, Id, _, _, _, _, _, _, _, _, _, _ = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")	
	return Id
end

function Capitalist:LinkExtractName(link)
	local _, _, _, _, _, _, _, _, _, _, _, _, _, _, Name = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
	return Name
end

function Capitalist:LinkStripSuffix(link)
	-- Example link:
	-- "|cff9d9d9d|Hitem:7073:0:0:0:0:0:0:0:80:0|h[Broken Fang]|h|r"

	local _, _, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging, Name = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

	local _, _, NameBase, NameSuffix = string.find(Name, "(.*) of the (.*)")

	-- replace the Suffix with zero and remove "of the ___" from the name
	-- Note: Only trim the "of the ___" if the item has a suffix so items with an original name of that form are left alone
	if (Suffix ~= "0") then
		Suffix = "0"
		if (NameSuffix) then
			Name = NameBase
		end
	end

	local linkStripped = ""
	if (Color) then
		linkStripped = linkStripped.."|cff"..Color
	end
	linkStripped = linkStripped.."|H"..Ltype
	linkStripped = linkStripped..":"..strjoin(":", Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging)
	if (Name) then
		linkStripped = linkStripped.."|h["..Name.."]|h|r"
	end

	return linkStripped
end

function Capitalist:LinkStripUniqueId(link)
	-- Example link:
	-- "|cff9d9d9d|Hitem:7073:0:0:0:0:0:0:0:80:0|h[Broken Fang]|h|r"

	local _, _, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging, Name = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")

	-- Generalize: replace the Unique with zero
	Unique = "0"

	local linkStripped = ""
	if (Color) then
		linkStripped = linkStripped.."|cff"..Color
	end
	linkStripped = linkStripped.."|H"..Ltype
	linkStripped = linkStripped..":"..strjoin(":", Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, reforging)
	if (Name) then
		linkStripped = linkStripped.."|h["..Name.."]|h|r"
	end

	return linkStripped
end
